from io import BytesIO
from typing import Dict, NamedTuple, Callable

import pytest
import requests
import json
import inspect
from platform import system

from pytest_httpserver import HTTPServer, HeaderValueMatcher
from pytest import mark
from requests_toolbelt import MultipartEncoder

from arthurai.client.base import BaseApiClient
from arthurai import __version__ as arthur_version


# CONSTANTS
from tests import MockResponse

MULTIPART_HEADER_PREFIX = "multipart/form-data; boundary="

# FIXTURES
# to be supplied
ACCESS_KEY = "abc123"
SAMPLE_DATA = {'foo': "bar",
               'baz': 2}
SAMPLE_DATA_BYTES = bytes(json.dumps(SAMPLE_DATA), encoding='utf-8')
NESTED_SAMPLE_DATA = {'data': json.dumps(SAMPLE_DATA)}
PARAMS = {'query': "param"}
SAMPLE_FILE_DATA = {
    "file.name": bytes("hello im mr data", encoding="utf-8"),
    "other.file": bytes("and im other data", encoding="utf-8"),
}

# to be returned
ORG_ID = "org123"
ORG_ID_RESPONSE = {'organization_id': ORG_ID}

# headers
INITIAL_HEADERS = {
    "Authorization": ACCESS_KEY,
    "Accept": "application/json"
}
JSON_HEADERS = {
    **INITIAL_HEADERS, "Content-Type": "application/json"}
MULTIPART_HEADERS = {
    **INITIAL_HEADERS, "Content-Type": "multipart/form-data"
}

EXPECTED_USER_AGENT = f"arthur-sdk/{arthur_version} (system={system()}, org={ORG_ID})"


def given_arthur_http_client(httpserver: HTTPServer) -> BaseApiClient:
    base_url = httpserver.url_for("")
    return BaseApiClient(access_key=ACCESS_KEY, url=base_url, base_path='/api/v3', verify_ssl=True)


"""
TIP:
The pytest_httpserver lib is generally great but a little annoying because if there's no match you just get back a 
generic "no matcher found" response. I suggest dropping a breakpoint in the httpserver.py/RequestMatcher.match()
method if you're trying to diagnose errors.
"""


class TestHTTPClientCase(NamedTuple):
    given_client: Callable
    user_agent: str
    api_prefix: str
    extra_requests_count: int
    multipart_requires_content_type: bool


CASES = [
    pytest.param(
        TestHTTPClientCase(given_client=given_arthur_http_client, user_agent=EXPECTED_USER_AGENT, api_prefix="",
                           extra_requests_count=1, multipart_requires_content_type=True), id="arthur-http-client")
]


@mark.parametrize("case", CASES)
def test_get(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # allow the client to include an application/json Content-Type header even though it shouldn't
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(INITIAL_HEADERS, case.user_agent),
                                      query_string=PARAMS) \
        .respond_with_json({'status': "ok"})
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      query_string=PARAMS) \
        .respond_with_json({'status': "ok"})

    response = client.get("/test", params=PARAMS, return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_delete(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # allow the client to include an application/json Content-Type header even though it shouldn't
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="DELETE",
                                      headers=add_user_agent_header(INITIAL_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="DELETE",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.delete("/test", return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_json_post(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST", json=SAMPLE_DATA,
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.post("/test", data=SAMPLE_DATA, return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_multipart_post(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': match_multipart_headers
    })

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    # our client makes you specify multipart in the headers yourself, otherwise it wont send the files you supply
    base_headers = {'Content-Type': "multipart/form-data"} if case.multipart_requires_content_type else {}
    response = client.post("/test", data=None, files=sample_files,
                           headers=add_user_agent_header(base_headers, case.user_agent),
                           return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map={})
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_multipart_post_retry(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'User-Agent': HeaderValueMatcher.default_header_value_matcher,
        'Content-Type': match_multipart_headers
    })

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'errors': "no bad"}, status=500)
    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="POST",
                                      header_value_matcher=multipart_header_matcher,
                                      headers=add_user_agent_header(MULTIPART_HEADERS, case.user_agent)) \
        .respond_with_json({'message': "ok"})

    # our client makes you specify multipart in the headers yourself, otherwise it wont send the files you supply
    base_headers = {'Content-Type': "multipart/form-data"} if case.multipart_requires_content_type else {}
    response = client.post("/test", data=None, files=sample_files,
                           headers=add_user_agent_header(base_headers, case.user_agent), retries=1,
                           return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map={})
    assert len(httpserver.log) == 2 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_json_patch(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="PATCH", json=SAMPLE_DATA,
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.patch("/test", data=SAMPLE_DATA, return_raw_response=True)

    assert response.status_code < 300
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_multipart_patch_mixed(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    sample_files = sample_file_data_to_files_map(SAMPLE_FILE_DATA)

    # header matchers, kinda weird API but this does the trick
    multipart_header_matcher = HeaderValueMatcher(matchers={
        **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
        'Content-Type': match_multipart_headers
    })

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="PATCH",
                                      header_value_matcher=multipart_header_matcher, headers=MULTIPART_HEADERS) \
        .respond_with_json({'status': "ok"})

    response = client.patch("/test", data=NESTED_SAMPLE_DATA, files=sample_files,
                            headers=add_user_agent_header({'Content-Type': "multipart/form-data"}, case.user_agent),
                            return_raw_response=True)

    assert response.status_code < 300
    request = httpserver.log[-1][0]
    assert_multipart_expected(request, files_map=sample_files, data_map={'data': SAMPLE_DATA_BYTES})
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_put(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    httpserver.expect_oneshot_request(f"{case.api_prefix}/test", method="PUT", json=SAMPLE_DATA,
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent)) \
        .respond_with_json({'status': "ok"})

    response = client.put("/test", data=SAMPLE_DATA, return_raw_response=True)

    assert response.status_code < 300
    # one for get org and one for current
    assert len(httpserver.log) == 1 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_retries_validation(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # headers
    def allow_none_match(actual: str, expected: str) -> bool:
        return (actual == expected) or actual is None
    allow_none_matcher = HeaderValueMatcher(matchers={
            **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
            'User-Agent': HeaderValueMatcher.default_header_value_matcher,
            'Content-Type': allow_none_match
        })

    # 400 response
    bad_request_response = MockResponse("bad request", 400)
    internal_error_response = MockResponse("internal server error", 500)

    if 'validation_response_code' in inspect.signature(client.send).parameters.keys():
        additional_args = {'validation_response_code': 200}
    else:
        additional_args = {}

    # two failures then one success
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(bad_request_response)
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(internal_error_response)
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_json({'status': "ok"})

    response = client.send("/test", method="GET", return_raw_response=True, retries=2, **additional_args)

    assert response.status_code < 300
    assert len(httpserver.log) == 3 + case.extra_requests_count


@mark.parametrize("case", CASES)
def test_retries_validation_failure(httpserver: HTTPServer, case: TestHTTPClientCase):
    given_orgid_returned(httpserver)
    client = case.given_client(httpserver)

    # headers
    def allow_none_match(actual: str, expected: str) -> bool:
        return (actual == expected) or actual is None
    allow_none_matcher = HeaderValueMatcher(matchers={
            **{header: HeaderValueMatcher.default_header_value_matcher for header in INITIAL_HEADERS.keys()},
            'User-Agent': HeaderValueMatcher.default_header_value_matcher,
            'Content-Type': allow_none_match
        })

    # 400 response
    bad_request_response = MockResponse("bad request", 400)
    internal_error_response = MockResponse("internal server error", 500)

    if 'validation_response_code' in inspect.signature(client.send).parameters.keys():
        additional_args = {'validation_response_code': 200}
    else:
        additional_args = {}

    # two failures then one success
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(bad_request_response)
    httpserver.expect_ordered_request(f"{case.api_prefix}/test", method="GET",
                                      headers=add_user_agent_header(JSON_HEADERS, case.user_agent),
                                      header_value_matcher=allow_none_matcher) \
        .respond_with_response(internal_error_response)

    exception = None
    response = None
    try:
        response = client.send("/test", method="GET", return_raw_response=True, retries=1, **additional_args)
    except Exception as e:
        exception = e

    assert exception is not None or (response is not None and response.status_code == 500)
    assert len(httpserver.log) == 2 + case.extra_requests_count


def add_user_agent_header(original_headers, user_agent):
    return {
        **original_headers,
        "User-Agent": user_agent
    }


def given_orgid_returned(httpserver: HTTPServer) -> None:
    # allow the client to include an application/json Content-Type header even though it shouldn't
    headers = {**INITIAL_HEADERS, "Content-Type": "application/json"}
    user_agent = f"python-requests/{requests.__version__}"
    httpserver.expect_oneshot_request("/api/v3/users/me", method="GET",
                                      headers=add_user_agent_header(headers, user_agent)) \
        .respond_with_json(ORG_ID_RESPONSE)
    httpserver.expect_oneshot_request("/api/v3/users/me", method="GET",
                                      headers=add_user_agent_header(INITIAL_HEADERS, user_agent)) \
        .respond_with_json(ORG_ID_RESPONSE)


def match_multipart_headers(actual: str, expected: str) -> bool:
    return actual.startswith("multipart/form-data; boundary=")


def sample_file_data_to_files_map(sample_data: Dict[str, bytes]) -> Dict[str, BytesIO]:
    return {
        fname: BytesIO(content) for fname, content in sample_data.items()
    }


def assert_multipart_expected(request: requests.Request, files_map: Dict[str, BytesIO],
                              data_map: Dict[str, bytes]) -> None:
    boundary = request.headers['Content-Type'][len(MULTIPART_HEADER_PREFIX):]
    prefix = bytes(f"--{boundary}", encoding="utf-8")

    data = request.data
    assert data is not None
    assert isinstance(data, bytes)
    assert data.startswith(prefix)

    blocks = data.split(prefix)

    # build map from expected content to count of instances
    block_counts_map = {}
    for filename, content in files_map.items():
        content.seek(0)
        expected_block = bytes(f'\r\nContent-Disposition: form-data; name="{filename}"; filename="{filename}"\r\n\r\n',
                               encoding='utf-8') + content.read() + bytes("\r\n", encoding='utf-8')
        block_counts_map[expected_block] = 0
    for keyname, data in data_map.items():
        expected_block = bytes(f'\r\nContent-Disposition: form-data; name="{keyname}"\r\n\r\n',
                               encoding='utf-8') + data + bytes("\r\n", encoding='utf-8')
        block_counts_map[expected_block] = 0

    # first block should be empty per startswith() assertion above, iterate through content blocks
    # go through each subsequent block, it should be in our map
    for block in blocks[1:-1]:
        assert block in block_counts_map.keys()
        block_counts_map[block] += 1

    # ensure each block was only seen once
    for count in block_counts_map.values():
        assert count == 1

    # validate ends with expected suffix
    assert blocks[-1] == bytes("--\r\n", encoding='utf-8')



import pytest 
import responses

import pandas as pd

from arthurai.client.base import BaseApiClient
from arthurai.core.models import ArthurModel
from arthurai.core.bias.threshold_mitigation import Curves

from .test_request_models.fixtures import model_response_json_strings
from .fixtures.mocks import BASE_URL, ACCESS_KEY

modelclient = BaseApiClient(url=BASE_URL, access_key=ACCESS_KEY, base_path="binarybasepath", offline=True)

BINARY_MODEL = ArthurModel.from_json(model_response_json_strings[4])
BINARY_MODEL._client = BaseApiClient(url=BASE_URL, access_key=ACCESS_KEY, base_path="binarybasepath", offline=True)

mock_curve_response = {
    'data': [ {
        'constraint': 'demographic_parity',
        'x_label': 'selection_rate',
        'y_label': 'accuracy',
        'model_id': 'test_modelid',
        'attribute_name': 'test_attrname',
        'categorical_value': 'attrval1', 
        'data_points': [
            {'x': 1, 'y': 2, 'threshold': 0.5}
        ],
        'optimization_index': 0
    }, 
    {
        'constraint': 'demographic_parity',
        'x_label': 'selection_rate',
        'y_label': 'accuracy',
        'model_id': 'test_modelid',
        'attribute_name': 'test_attrname',
        'categorical_value': 'attrval2', 
        'data_points': [
            {'x': 1.5, 'y': 2.5, 'threshold': 0.55}
        ],
        'optimization_index': 0
    }
    ]
}

mock_curve_response_cont = {
    'data': [ {
        'constraint': 'demographic_parity',
        'x_label': 'selection_rate',
        'y_label': 'accuracy',
        'model_id': 'test_modelid',
        'attribute_name': 'test_attrname',
        'continuous_end': 1, 
        'data_points': [
            {'x': 1, 'y': 2, 'threshold': 0.5}
        ],
        'optimization_index': 0
    }, 
    {
        'constraint': 'demographic_parity',
        'x_label': 'selection_rate',
        'y_label': 'accuracy',
        'model_id': 'test_modelid',
        'attribute_name': 'test_attrname',
        'continuous_start': 1,
        'continuous_end': 2,
        'data_points': [
            {'x': 1.5, 'y': 2.5, 'threshold': 0.55}
        ],
        'optimization_index': 0
    },
    {
        'constraint': 'demographic_parity',
        'x_label': 'selection_rate',
        'y_label': 'accuracy',
        'model_id': 'test_modelid',
        'attribute_name': 'test_attrname',
        'continuous_start': 2,
        'data_points': [
            {'x': 1.1, 'y': 2.1, 'threshold': 0.51}
        ],
        'optimization_index': 0
    }
    ]
}

sampledict = {
    'val_1': pd.DataFrame({'x': [1, 2, 3], 'y': [2,4,6], 'threshold': [0.1, 0.2, 0.3]}),
    'val_2': pd.DataFrame({'x': [1, 2, 3], 'y': [2,4,6], 'threshold': [0.25, 0.35, 0.45]}),
}

test_curve = Curves(attribute_name="testattrname", constraint="equalized_odds", max_acc_idx=1, attr_val_to_threshold = sampledict )

@responses.activate
def test_get_curves():

    responses.add(responses.GET, "http://mockbinarybasepath/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasbinary/enrichments/bias_mitigation/curves?attribute_id=8514e278-24d1-4e20-b209-3c77eb2b247f&constraint=demographic_parity",
        json = mock_curve_response)
    
    mitigator = BINARY_MODEL.bias.mitigation_threshold 
    curves = mitigator.get_curves("attr_1", "demographic_parity")

    assert curves.constraint == "demographic_parity"
    assert curves.max_acc_idx == 0
    assert curves.attr_val_to_threshold["attrval2"].iloc[0]['x'] == 1.5


@responses.activate
def test_get_curves_continuous():

    responses.add(responses.GET, "http://mockbinarybasepath/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasbinary/enrichments/bias_mitigation/curves?attribute_id=8514e278-24d1-4e20-b209-3c77eb2b247f&constraint=demographic_parity",
        json = mock_curve_response_cont)
    
    mitigator = BINARY_MODEL.bias.mitigation_threshold 
    curves = mitigator.get_curves("attr_1", "demographic_parity")

    assert curves.constraint == "demographic_parity"
    assert curves.max_acc_idx == 0
    assert '(-inf,1]' in curves.attr_val_to_threshold.keys()
    assert '(1,2]' in curves.attr_val_to_threshold.keys()
    assert curves.attr_val_to_threshold["(2,inf)"].iloc[0]['x'] == 1.1

def test_get_thresholds_for_idx_default():
    mitigator = BINARY_MODEL.bias.mitigation_threshold 

    idxs = mitigator.get_thresholds_for_idx(test_curve)

    assert idxs['val_1'] == 0.2
    assert idxs['val_2'] == 0.35
    assert len(idxs.keys()) == 2

def test_get_thresholds_for_idx_spec():
    mitigator = BINARY_MODEL.bias.mitigation_threshold 

    idxs = mitigator.get_thresholds_for_idx(test_curve, idx=0)

    assert idxs['val_1'] == 0.1
    assert idxs['val_2'] == 0.25
    assert len(idxs.keys()) == 2


import logging
import os
import tempfile
from math import ceil
from typing import Dict, Optional, Any, TYPE_CHECKING, List

from pandas import DataFrame
import pandas as pd
import zipfile
import shutil
from pathlib import PurePath, Path
import uuid

from arthurai.common.constants import Stage, InputType
from arthurai.common.exceptions import UserValueError
from arthurai.core import util
#  imports ArthurModel for type checking, required due to circular import
if TYPE_CHECKING:
    from arthurai.core.models import ArthurModel

logger = logging.getLogger(__name__)

# fixed constants
MEGABYTE_SIZE = 1024 * 1024

# config constants
ROW_GROUP_SIZE = 5_000
# TODO: a more robust approach would be to write a single row group, measure the file size, and use that to decide how
#  big to make the files
MAX_ROWS_PER_FILE = 500_000


class DatasetService:
    COUNTS = "counts"
    SUCCESS = "success"
    FAILURE = "failure"
    TOTAL = "total"
    FAILURES = "failures"
    DEFAULT_MAX_IMAGE_DATA_BYTES = 300000000  # 300 MB

    @staticmethod
    def convert_dataframe(model_id: str, stage: Optional[Stage], df: DataFrame,
                          max_rows_per_file=MAX_ROWS_PER_FILE) -> str:
        """Convert a dataframe to parquet named {model.id}-{stage}.parquet in the system tempdir

        :param model_id: a model id
        :param stage: the :py:class:`.Stage`
        :param df: the dataframe to convert
        :param max_rows_per_file: the maximum number of rows per parquet file

        Returns:
            The filename of the parquet file that was created
        """
        temp_dir = tempfile.mkdtemp()
        num_chunks = ceil(len(df) / max_rows_per_file)
        for chunk in range(num_chunks):
            fname = model_id
            fname += f"-{stage}" if stage is not None else ""
            fname += f"-{chunk}.parquet"
            filename = os.path.join(temp_dir, fname)
            start_idx = chunk * max_rows_per_file
            end_idx = (chunk + 1) * max_rows_per_file
            df.iloc[start_idx:end_idx].to_parquet(filename, index=False, allow_truncated_timestamps=True,
                                                  row_group_size=ROW_GROUP_SIZE, engine="pyarrow")
        return temp_dir
    
    @staticmethod
    def chunk_parquet_image_set(directory_path: str, image_attribute: str,
                                max_image_data_bytes: int = DEFAULT_MAX_IMAGE_DATA_BYTES) -> str:
        """Takes in a directory path with parquet files containing image attributes.
        Divides images up into 300MB chunks, then zipped, the parquet file is also split up to match.
        Parquet files will have random filename, and image zips will have matching name.

        """
        # make output dir for storing all chunks. At end will get:
        # tmp_dir/
        #    123.parquet, 123.zip, 456.parquet, 456.zip
        # TODO remove print statements, add logs to indicate processing, can take a while to run
        output_dir = tempfile.mkdtemp()
        files = util.retrieve_parquet_files(directory_path)
        if not files:
            raise UserValueError("The directory supplied does not contain any parquet files to upload")
        
        # loop through each parquet file
        for file in files:
            # keep track of where we are in parquet file, in case file needs to be split
            # to match image chunk
            cur_size = 0
            last_df_chunk_index = 0
            cur_img_dir = tempfile.mkdtemp(prefix=output_dir + "/")

            # this shouldn't happen, but check just in case
            if file.suffix != ".parquet":
                continue
            df = pd.read_parquet(file)

            if image_attribute not in df:
                # TODO should we raise exception here instead?
                logger.warning(f"Found parquet file missing image attribute, not including in reference set: {file.name}")
                continue
            
            # loop through each row in parquet file
            for cur_df_index, image_path in enumerate(df[image_attribute]):
                # verify image exists
                if not os.path.exists(image_path):
                    # TODO raise error here?
                    logger.warning(f"Image does not exist for row, not including in reference set: {image_path}")
                    continue
                
                # move image to temp dir
                image_path = PurePath(image_path)
                temp_image_path = os.path.join(cur_img_dir, image_path.name)
                shutil.copyfile(image_path, temp_image_path)
                img_bytes = os.path.getsize(temp_image_path)
                cur_size += img_bytes

                # if we have reached max image file size, save and start new chunk
                if cur_size >= max_image_data_bytes:
                    chunk_name = str(uuid.uuid4())

                    # create parquet chunk
                    df_chunk = df.iloc[last_df_chunk_index:cur_df_index + 1]
                    # replace image attribute with just the filename, no path
                    df_chunk[image_attribute] = df_chunk[image_attribute].apply(lambda x: PurePath(x).name)
                    df_chunk_filename = f"{chunk_name}.parquet"
                    df_chunk_path = os.path.join(output_dir, df_chunk_filename)
                    df_chunk.to_parquet(df_chunk_path)

                    # zip images
                    image_zip_path = os.path.join(output_dir, chunk_name)
                    shutil.make_archive(image_zip_path, 'zip', cur_img_dir)

                    # reset for next chunk
                    shutil.rmtree(cur_img_dir)
                    cur_img_dir = tempfile.mkdtemp(prefix=output_dir + "/")
                    cur_size = 0
                    last_df_chunk_index = cur_df_index + 1
            # we have reached end of current file, close off the current chunk before next parquet file
            # TODO maybe pull this into function so no repeated code, but so many things to pass in
            chunk_name = str(uuid.uuid4())

            # create parquet chunk
            df_chunk = df.iloc[last_df_chunk_index:cur_df_index + 1]
            # replace image attribute with just the filename, no path
            df_chunk[image_attribute] = df_chunk[image_attribute].apply(lambda x: PurePath(x).name)
            df_chunk_filename = f"{chunk_name}.parquet"
            df_chunk_path = os.path.join(output_dir, df_chunk_filename)
            df_chunk.to_parquet(df_chunk_path)

            # zip images
            image_zip_path = os.path.join(output_dir, chunk_name)
            shutil.make_archive(image_zip_path, 'zip', cur_img_dir)

            # clean up
            shutil.rmtree(cur_img_dir)
        return output_dir

    @staticmethod
    def files_size(parquet_files: List[Path], model_input_type: InputType) -> int:
        all_files = parquet_files.copy()

        # extra image zip file should have same path and name as parquet file if model is image model
        if model_input_type == InputType.Image:
            for pf in parquet_files:
                all_files.append(Path(os.path.join(pf.parent, pf.stem) + ".zip"))
        total_size_bytes = 0
        for file_path in all_files:
            try:
                total_size_bytes += os.path.getsize(file_path)
            except FileNotFoundError:
                pass
        return total_size_bytes

    @staticmethod
    def send_parquet_files_from_dir_iteratively(model: 'ArthurModel', directory_path: str,
                                                url: str, upload_file_param_name: str,
                                                additional_form_params: Optional[Dict[str, Any]] = None,
                                                retries: int = 0):
        """Sends parquet files iteratively from a specified directory to a specified url for a given model

        :param retries: Number of times to retry the request if it results in a 400 or higher response code
        :param model:    the :py:class:`!arthurai.client.apiv2.model.ArthurModel`
        :param directory_path:    local path containing parquet files to send
        :param url:    POST url endpoint to send files to
        :param upload_file_param_name:     name to use in body with attached files
        :param additional_form_params: dictionary of additional form file params to send along with parquet file

        :raises MissingParameterError: the request failed

        :returns A list of files which failed to upload
        """
        files = util.retrieve_parquet_files(directory_path)
        if not files:
            raise UserValueError("The directory supplied does not contain any parquet files to upload")

        total_size = DatasetService.files_size(files, model.input_type)
        logger.info(f"Starting upload ({total_size/MEGABYTE_SIZE:.3f} MB in {len(files)} files), depending on data "
                    f"size this may take a few minutes")

        failed_files = []
        succeeded_files = []
        expected_keys = {DatasetService.SUCCESS, DatasetService.FAILURE, DatasetService.TOTAL}

        counts = {
            DatasetService.SUCCESS: 0,
            DatasetService.FAILURE: 0,
            DatasetService.TOTAL: 0
        }
        failures: List[Any] = []

        for file in files:
            if file.suffix == '.parquet':
                with open(file, 'rb') as parquet_file:
                    headers = {'Content-Type': 'multipart/form-data'}
                    form_parts = {} if additional_form_params is None else additional_form_params
                    form_parts.update({upload_file_param_name: parquet_file})

                    # add corresponding image data if image model
                    if model.input_type == InputType.Image:
                        # image zip file has same path and name as parquet file
                        image_zip_name = str(os.path.join(file.parent, file.stem)) + ".zip"
                        image_zip_file = open(image_zip_name, 'rb')
                        form_parts.update({'image_data': (image_zip_name, image_zip_file, "application/zip")})

                    resp = model._client.post(url, data=None, files=form_parts, headers=headers,
                                              return_raw_response=True, retries=retries)
                    if resp.status_code == 201:
                        logger.info(f"Uploaded completed: {file}")
                        succeeded_files.append(file)
                    elif resp.status_code == 207:
                        logger.info(f"Upload completed: {file}")
                        result: Dict[str, Dict[str, int]] = resp.json()
                        # ensure the response is in the correct format
                        if DatasetService.COUNTS in result and DatasetService.FAILURES in result \
                                and set(result[DatasetService.COUNTS].keys()) == expected_keys:
                            counts[DatasetService.SUCCESS] += \
                                result[DatasetService.COUNTS][DatasetService.SUCCESS]
                            counts[DatasetService.FAILURE] += \
                                result[DatasetService.COUNTS][DatasetService.FAILURE]
                            counts[DatasetService.TOTAL] += \
                                result[DatasetService.COUNTS][DatasetService.TOTAL]
                            failures.append(result[DatasetService.FAILURES])
                        else:
                            failures.append(result)
                    else:
                        logger.error(f"Failed to upload file: {resp.text}")
                        failed_files.append(file)
                        failures.append(resp.json())
                        counts[DatasetService.FAILURE] += 1
                        counts[DatasetService.TOTAL] += 1
                # close image zip
                if model.input_type == InputType.Image:
                    image_zip_file.close()
                    try:
                        os.remove(image_zip_file.name)
                    except Exception:
                        logger.warning(f"Failed to delete temporary image file at {image_zip_file.name}")
                    

        file_upload_info = {
            DatasetService.COUNTS: counts,
            DatasetService.FAILURES: failures
        }

        # Only log failed or succeeded files if they exist
        if len(failed_files) > 0:
            logger.error(f'Failed to upload {len(failed_files)} files')
        if len(succeeded_files) > 0:
            logger.info(f'Successfully uploaded {len(succeeded_files)} files')
        return failed_files, file_upload_info


class ImageZipper:

    def __init__(self):
        self.temp_file = tempfile.NamedTemporaryFile()
        self.zip = zipfile.ZipFile(self.temp_file.name, 'w')

    def add_file(self, path: str):
        self.zip.write(path)

    def get_zip(self):
        self.zip.close()
        return self.temp_file

    def __del__(self):
        self.zip.close()
        self.temp_file.close()



import json
from typing import Union, Dict, List
from requests import Response

import pytest
from unittest import mock

from arthurai.client.validation import validate_multistatus_response_and_get_failures





import responses
import pytest

from arthurai.client.base import BaseApiClient
from arthurai.core.models import ArthurModel
from arthurai.core.bias.bias_metrics import BiasMetrics

from .test_request_models.fixtures import model_response_json_strings
from .fixtures.mocks import BASE_URL, ACCESS_KEY

modelclient = BaseApiClient(url=BASE_URL, access_key=ACCESS_KEY, base_path="binarybasepath", offline=True)

BINARY_MODEL = ArthurModel.from_json(model_response_json_strings[4])
BINARY_MODEL._client = BaseApiClient(url=BASE_URL, access_key=ACCESS_KEY, base_path="binarybasepath", offline=True)
MULTICLASS_MODEL = ArthurModel.from_json(model_response_json_strings[5])
MULTICLASS_MODEL._client = BaseApiClient(url=BASE_URL, access_key=ACCESS_KEY, base_path="multiclassbasepath", offline=True)

bias_query_json_result = {
            "query_result": [
            {
                "bias": 1,
                "confusion_matrix": {
                "accuracy_rate": 0.9,
                "balanced_accuracy_rate": 0.8,
                "f1": 0.7,
                "false_negative_rate": 0.5,
                "false_positive_rate": 0,
                "precision": 0.9,
                "true_negative_rate": 1,
                "true_positive_rate": 0.5
                }
            },
            {
                "bias": 2,
                "confusion_matrix": {
                "accuracy_rate": 0.9,
                "balanced_accuracy_rate": 0.8,
                "f1": 0.7,
                "false_negative_rate": 0.4,
                "false_positive_rate": 0,
                "precision": 0.8,
                "true_negative_rate": 1,
                "true_positive_rate": 0.6
                }
            }
            ],
            "sampling_threshold": 1
        }
    

@responses.activate
def test_get_bias_cm_binary():

    responses.add(responses.POST, "http://mockbinarybasepath/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasbinary/inferences/query",
        json = bias_query_json_result
    )

    bin_metric = BiasMetrics(BINARY_MODEL)

    metrics = bin_metric.group_confusion_matrices("bias")
    metrics_flipped = bin_metric.group_confusion_matrices("bias", return_by_metric=False)

    assert "precision" in metrics.keys()
    assert len(metrics.keys()) == 8
    assert len(metrics['f1'].keys()) == 2
    assert len(metrics_flipped[1].keys()) == 8
    assert metrics_flipped[1]['accuracy_rate'] == 0.9

@responses.activate
def test_get_bias_cm_multiclass():
    responses.add(responses.POST, "http://mockmulticlassbasepath/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasmulticlass/inferences/query",
        json = bias_query_json_result
    )

    mult_metric = BiasMetrics(MULTICLASS_MODEL)
    metrics = mult_metric.group_confusion_matrices("bias", "attr_2")
    metrics_flipped = mult_metric.group_confusion_matrices("bias", "attr_2", return_by_metric=False)

    assert "precision" in metrics.keys()
    assert len(metrics.keys()) == 8
    assert len(metrics['f1'].keys()) == 2
    assert len(metrics_flipped[1].keys()) == 8
    assert metrics_flipped[1]['accuracy_rate'] == 0.9

def test_get_bias_cm_multiclass_requires_predvalue():
    mult_metric = BiasMetrics(MULTICLASS_MODEL)
    with pytest.raises(Exception):
        mult_metric.group_confusion_matrices("bias")

@responses.activate
def test_get_bias_pr_binary():
    responses.add(responses.POST, "http://mockbinarybasepath/models/ac55c7b4-2db7-4902-8cc3-969ed67a20c8_biasbinary/inferences/query",
        json = {
            "query_result": [
            {
                "bias": 1,
                "positive_rate": 0.15
            },
            {
                "bias": 2,
                "positive_rate": 0.14
            }
            ],
            "sampling_threshold": 1
        }
    )

    bin_metric = BiasMetrics(BINARY_MODEL)
    prs = bin_metric.group_positivity_rates("bias")
    prsalias = bin_metric.demographic_parity("bias")

    assert prs == prsalias    
    assert len(prs.keys()) == 2
    assert prs[2] == 0.14


def test_get_bias_pr_multiclass():
    mult_metric = BiasMetrics(MULTICLASS_MODEL)
    with pytest.raises(Exception):
        mult_metric.group_positivity_rates("bias")


from datetime import datetime, timezone
import numpy as np
import pandas as pd

from mock import patch, mock_open, MagicMock
import pytest
import responses

from tests.fixtures.models import regression as reg, multiclass as mclass
from tests.fixtures.models.regression import initial_batch_model as reg_initial_batch_model, \
    unsaved_batch_model as reg_unsaved_batch_model
from tests.fixtures.models.multiclass import initial_biclass_model
from tests.fixtures.mocks import client

from arthurai.common.constants import ValueType, Stage, InputType, OutputType, ImageResponseType, ImageContentType
from arthurai.core.attributes import AttributeCategory, AttributeBin, ArthurAttribute
from arthurai.core.models import ArthurModel
import arthurai.core.util as core_util
from arthurai.common.exceptions import UserValueError, MethodNotApplicableError, ArthurInternalError
from tests.helpers import mock_get, assert_attributes_equal, mock_post
from tests.test_request_models.fixtures import model_response_json_strings


@responses.activate
def test_send_parquet_file():
    input_dict = {
        "attr1": np.Inf,
        "attr2": "string1",
        "attr3": 3.44,
        "attr4": True,
        "attr5": np.nan,
        "attr6": -np.inf,
        "attr7": "string2",
        "attr8": None,
        "attr9": "",
        "attr10": np.inf,
    }
    expected_dict = {
        "attr1": None,
        "attr2": "string1",
        "attr3": 3.44,
        "attr4": True,
        "attr5": None,
        "attr6": None,
        "attr7": "string2",
        "attr8": None,
        "attr9": "",
        "attr10": None,
    }
    output_dict = ArthurModel._replace_nans_and_infinities_in_dict(input_dict)
    assert expected_dict == output_dict


def test_replace_nans_none_type():
    output_dict = ArthurModel._replace_nans_and_infinities_in_dict(None)
    assert output_dict is None


def test_standardize_pd_obj_for_nans():
    input_df = pd.DataFrame(
        [
            [1, 2, 3., 4, 5, 6, np.inf, np.nan, 1, 1],
            [7, 8, 9., 1, 2, 3, np.nan, -np.inf, 2, 2],
            [4, 5.5, 9., -np.inf, np.nan, np.Inf, np.nan, np.nan, np.nan, np.nan],
        ],
        columns = ["x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9"]
    )
    expected_df = pd.DataFrame(
        [
            [1, 2, 3., 4, 5, 6, np.nan, np.nan, 1, 1],
            [7, 8, 9., 1, 2, 3, np.nan, np.nan, 2, 2],
            [4, 5.5, 9., np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan],
        ],
        columns = ["x0", "x1", "x2", "x3", "x4", "x5", "x6", "x7", "x8", "x9"]
    )
    expected_df.x3 = expected_df.x3.astype(pd.Int64Dtype())
    expected_df.x4 = expected_df.x4.astype(pd.Int64Dtype())
    expected_df.x5 = expected_df.x5.astype(pd.Int64Dtype())

    attributes = {
        "x3": ValueType.Integer,
        "x4": ValueType.Integer,
        "x5": ValueType.Integer,
        "x8": ValueType.Float,
    }

    assert expected_df.equals(
        core_util.standardize_pd_obj(input_df, dropna=False, replacedatetime=False, attributes=attributes)
    )
    with pytest.raises(ValueError):
        assert expected_df.equals(
            core_util.standardize_pd_obj(input_df, dropna=True, replacedatetime=False, attributes=attributes)
        )


def test_standardize_pd_obj_for_series():
    input_series = [
        pd.Series([1, 2, 3]),
        pd.Series([1., 2., 3.]),
        pd.Series([1, 2, np.Inf], name="x1"),
        pd.Series([1, 2, np.Inf], name="x1"),
        pd.Series([1, 2, np.Inf], name="x1"),
        pd.Series([np.nan, -np.inf, np.inf]),
    ]
    input_attributes = [
        {},
        None,
        {"x1": ValueType.Integer},
        {"x1": ValueType.Float},
        {"x2": ValueType.Integer},
        {}
    ]
    expected_undropped_series = [
        pd.Series([1, 2, 3]),
        pd.Series([1., 2., 3.]),
        pd.Series([1, 2, np.nan]).astype(pd.Int64Dtype()),
        pd.Series([1, 2, np.nan]),
        pd.Series([1, 2, np.nan]),
        pd.Series([np.nan, np.nan, np.nan]),
    ]
    expected_dropped_series = [
        pd.Series([1, 2, 3]),
        pd.Series([1., 2., 3.]),
        pd.Series([1, 2]).astype(pd.Int64Dtype()),
        pd.Series([1, 2.]),
        pd.Series([1, 2.]),
        pd.Series([]),
    ]

    for inp, inp_attributes, exp_undropped, exp_dropped in zip(input_series, input_attributes, expected_undropped_series, expected_dropped_series):
        print(inp_attributes)
        assert exp_undropped.equals(
            core_util.standardize_pd_obj(inp, dropna=False, replacedatetime=False, attributes=inp_attributes)
        )
        assert exp_dropped.equals(
            core_util.standardize_pd_obj(inp, dropna=True, replacedatetime=False, attributes=inp_attributes)
        )


def test_standardize_pd_obj_bad_input():
    for inp in ["string", "", 123, [1,2,3], {"key": "value"}]:
        with pytest.raises(TypeError):
            core_util.standardize_pd_obj(inp, dropna=True, replacedatetime=False)


def test_standardize_pd_obj_for_datetimes():
    unaware_timestamp = datetime(2021, 2, 8, 12, 10, 5)
    aware_timestamp = datetime(2021, 2, 8, 12, 10, 5, tzinfo=timezone.utc)
    timestamp_to_expected_output = (
        (unaware_timestamp, ValueError),
        (aware_timestamp, "2021-02-08T12:10:05Z"),
    )

    for timestamp, expected_output in timestamp_to_expected_output:
        if isinstance(expected_output, str):
            actual_output = core_util.standardize_pd_obj(pd.Series([timestamp]), dropna=False, replacedatetime=True)
            assert expected_output == actual_output.values[0]
        else:
            with pytest.raises(expected_output):
                core_util.standardize_pd_obj(pd.Series([timestamp]), dropna=False, replacedatetime=True)


def test_add_attribute_with_categories():
    model = ArthurModel.from_json(model_response_json_strings[2])
    model.add_attribute(
        name="test_categorical_1",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        categories=[1, 2, 3, 4],
    )
    attr = model.get_attribute(name="test_categorical_1")
    assert len(attr.categories) == 4
    assert isinstance(attr.categories[0], AttributeCategory)

    model.add_attribute(
        name="test_categorical_2",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        categories=[AttributeCategory(value="hello")],
    )
    attr = model.get_attribute(name="test_categorical_2")
    assert len(attr.categories) == 1
    assert isinstance(attr.categories[0], AttributeCategory)
    assert attr.categories[0].value == "hello"


def test_add_attribute_with_bins():
    model = ArthurModel.from_json(model_response_json_strings[2])
    model.add_attribute(
        name="test_bins_1",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        bins=[None, 10, 15, 20, None],
    )
    attr = model.get_attribute(name="test_bins_1")
    assert len(attr.bins) == 4
    assert isinstance(attr.bins[0], AttributeBin)
    assert attr.bins[0].continuous_start is None
    assert attr.bins[0].continuous_end == 10

    model.add_attribute(
        name="test_bins_2",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        bins=[
            AttributeBin(continuous_start=None, continuous_end=10),
            AttributeBin(continuous_start=10, continuous_end=15),
            AttributeBin(continuous_start=15, continuous_end=20),
        ],
    )
    attr = model.get_attribute(name="test_bins_2")
    assert len(attr.bins) == 3
    assert isinstance(attr.bins[0], AttributeBin)
    assert attr.bins[0].continuous_start is None
    assert attr.bins[0].continuous_end == 10


def test_add_nlp_attribute():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Multiclass,
    )
    ref_data = pd.DataFrame(
        {"input_text": ["hello", "hi", "what's up", "yo", "cool"],}
    )
    model.infer_schema(ref_data, stage=Stage.ModelPipelineInput)

    assert (
        model.get_attribute("input_text").value_type == ValueType.Unstructured_Text
    )
    assert model.get_attribute("input_text").categories is None


def test_all_null_column():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Multiclass,
    )
    df = pd.DataFrame({'col1': [1], 'col2': [float("nan")]})
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='1', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.Float, stage=Stage.ModelPipelineInput, position=1,
                        categorical=False, is_unique=False)]

    assert model.attributes == expected_attributes


def test_nonzero_index_column():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Multiclass,
    )
    df = pd.DataFrame({'col1': [1], 'col2': [float("nan")]})
    df.index = [5]
    model.infer_schema(df, stage=Stage.ModelPipelineInput)

    expected_attributes = [
        ArthurAttribute(name='col1', value_type=ValueType.Integer, stage=Stage.ModelPipelineInput, position=0,
                        categorical=True, categories=[AttributeCategory(value='1', label=None)], is_unique=False),
        ArthurAttribute(name='col2', value_type=ValueType.Float, stage=Stage.ModelPipelineInput, position=1,
                        categorical=False, is_unique=False)]

    assert model.attributes == expected_attributes


def test_set_categorical_value_labels():
    model = ArthurModel.from_json(model_response_json_strings[2])
    model.add_attribute(
        name="test_categorical_1",
        value_type=ValueType.Integer,
        stage=Stage.ModelPipelineInput,
        categorical=True,
        categories=[1, 2],
    )
    expected_categories = {1: "Male", 2: "Female"}
    model.set_attribute_labels("test_categorical_1", labels=expected_categories)
    cats = model.get_attribute("test_categorical_1").categories
    for category in cats:
        assert int(category.value) in expected_categories
        assert expected_categories[int(category.value)] == category.label


@responses.activate
def test_get_image(client):
    model = ArthurModel(client=client.client,
                        partner_model_id="test",
                        input_type=InputType.Image,
                        output_type=OutputType.Multiclass)
    model.id = '1234567890abcdef'
    image_id = '0a1b2c3d4e5f9687'
    path = '.'
    type = ImageResponseType.RawImage
    file_ext = '.png'
    response_content = 'content'.encode()
    expected_image_file = f"{path}/{type}_{image_id}{file_ext}"

    mock_get(f"/api/v3/models/{model.id}/inferences/images/{image_id}?type={type}", response_content,
             headers={'Content-Type': ImageContentType.Png})
    open_mock = mock_open()
    with patch("arthurai.core.models.open", open_mock, create=True):
        resulting_file = model.get_image(image_id, path, type=type)

    assert resulting_file == expected_image_file
    open_mock.assert_called_with(expected_image_file, 'wb')
    open_mock.return_value.write.assert_called_once_with(response_content)


def test_add_object_detection_output_attributes():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Image,
        output_type=OutputType.Multiclass,
    )

    # cannot call without proper types
    with pytest.raises(MethodNotApplicableError):
        model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])
    model.input_type = InputType.NLP
    model.output_type = OutputType.ObjectDetection
    with pytest.raises(MethodNotApplicableError):
        model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])

    # cannot call without classes
    model.output_type = OutputType.ObjectDetection
    model.input_type = InputType.Image
    with pytest.raises(UserValueError):
        model.add_object_detection_output_attributes('pred', 'gt', [])

    # names cannot match
    with pytest.raises(UserValueError):
        model.add_object_detection_output_attributes('pred', 'pred', ['class_1', 'class_2'])

    # happy path
    model.add_object_detection_output_attributes('pred', 'gt', ['class_1', 'class_2'])
    assert model.get_attribute('pred').value_type == ValueType.BoundingBox
    assert model.get_attribute('gt').value_type == ValueType.BoundingBox
    assert model.image_class_labels == ['class_1', 'class_2']


def test_build_tabular_model(reg_initial_batch_model):
    pred_gt_map = {reg.PRED: reg.GROUND_TRUTH}
    reg_initial_batch_model.build(data=reg.SAMPLE_DATAFRAME, pred_to_ground_truth_map=pred_gt_map,
                                  non_input_columns=[reg.FLOAT_NONINPUT])

    assert_attributes_equal(reg_initial_batch_model.attributes, reg.MODEL_ATTRIBUTES)
    pd.testing.assert_frame_equal(reg_initial_batch_model.reference_dataframe, reg.SAMPLE_DATAFRAME)


def test_build_tabular_class_model(initial_biclass_model):
    pred_gt_map = {mclass.DOG_PRED: mclass.DOG_GROUND_TRUTH, mclass.CAT_PRED: mclass.CAT_GROUND_TRUTH}
    initial_biclass_model.build(data=mclass.SAMPLE_DATAFRAME, pred_to_ground_truth_map=pred_gt_map,
                                non_input_columns=[mclass.FLOAT_NONINPUT], positive_predicted_attr=mclass.DOG_PRED)

    assert_attributes_equal(initial_biclass_model.attributes, mclass.MODEL_ATTRIBUTES)
    pd.testing.assert_frame_equal(initial_biclass_model.reference_dataframe, mclass.SAMPLE_DATAFRAME)


def test_build_tabular_model_no_refset(reg_unsaved_batch_model):
    pred_gt_map = {reg.PRED: reg.GROUND_TRUTH}
    reg_unsaved_batch_model.build(data=reg.SAMPLE_DATAFRAME, pred_to_ground_truth_map=pred_gt_map,
                                  non_input_columns=[reg.FLOAT_NONINPUT], set_reference_data=False)

    assert_attributes_equal(reg_unsaved_batch_model.attributes, reg.MODEL_ATTRIBUTES)
    assert reg_unsaved_batch_model.reference_dataframe is None


def test_build_nlp():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.NLP,
        output_type=OutputType.Regression,
    )
    ref_data = pd.DataFrame({"input_text": ["hello there", "hi", "what's up", "yo", "cool"],
                             "pred": [90, 80, 50, 31, 26],
                             "gt": [95, 81, 42, 30, 30]})
    model.build(ref_data, pred_to_ground_truth_map={"pred": "gt"})

    assert model.get_attribute("input_text").value_type == ValueType.Unstructured_Text
    assert model.get_attribute("input_text").categories is None
    pd.testing.assert_frame_equal(model.reference_dataframe, ref_data)


def test_build_image():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Image,
        output_type=OutputType.Regression,
    )
    data = pd.DataFrame({'image': ["path1", "path2"], 'preds': [15.2, 31.6], 'gt': [12, 40]})
    model.build(data, pred_to_ground_truth_map={'preds': 'gt'})

    assert model.get_attribute("image").value_type == ValueType.Image
    assert model.get_attribute("image").categories is None
    pd.testing.assert_frame_equal(model.reference_dataframe, data)


def test_build_object_detection():
    model = ArthurModel(
        partner_model_id="test",
        client=None,
        input_type=InputType.Image,
        output_type=OutputType.ObjectDetection,
    )
    # not actually valid object data but that's okay
    data = pd.DataFrame({'image': ["path1", "path2"], 'preds': [0.8, 0.3], 'gt': [1, 0]})
    with pytest.raises(MethodNotApplicableError):
        model.build(data, pred_to_ground_truth_map={'preds': 'gt'})


@responses.activate
def test_save_model_with_refset(reg_unsaved_batch_model):
    reg_unsaved_batch_model.reference_dataframe = reg.SAMPLE_DATAFRAME
    mock_set_ref_data = MagicMock()
    reg_unsaved_batch_model.set_reference_data = mock_set_ref_data
    mock_post("/api/v3/models", response_body={'id': reg.TABULAR_MODEL_ID}, status=201)

    reg_unsaved_batch_model.save()

    mock_set_ref_data.assert_called_once_with(data=reg.SAMPLE_DATAFRAME)
    assert len(responses.calls) == 1


@responses.activate
def test_save_model_no_refset(reg_unsaved_batch_model):
    mock_set_ref_data = MagicMock()
    reg_unsaved_batch_model.set_reference_data = mock_set_ref_data
    mock_post("/api/v3/models", response_body={'id': reg.TABULAR_MODEL_ID}, status=201)

    reg_unsaved_batch_model.save()

    mock_set_ref_data.assert_not_called()
    assert len(responses.calls) == 1


def test_add_attribute_positions(reg_unsaved_batch_model):
    reg_unsaved_batch_model.infer_schema(reg.SAMPLE_DATAFRAME[reg.INT_INPUT], stage=Stage.ModelPipelineInput)
    reg_unsaved_batch_model.infer_schema(reg.SAMPLE_DATAFRAME[reg.FLOAT_NONINPUT], stage=Stage.NonInputData)
    reg_unsaved_batch_model.add_regression_output_attributes({reg.PRED: reg.GROUND_TRUTH}, value_type=ValueType.Integer)
    # there is one actual attribute in each stage so should be two in each with duplicates
    expected_attr_positions = {0, 1}
    for stage in (Stage.ModelPipelineInput, Stage.NonInputData, Stage.PredictedValue, Stage.GroundTruth):
        actual_attr_positions = [attr.position for attr in reg_unsaved_batch_model.attributes if attr.stage == stage]
        assert len(actual_attr_positions) == len(expected_attr_positions)
        assert set(actual_attr_positions) == expected_attr_positions
